"""
Thread Updater module to help update GUI items in a separate thread.
"""
from .__meta__ import version as __version__

from qt_thread_updater.thread_updater import is_main_thread, ThreadUpdater
from qt_thread_updater.global_utils import get_updater, set_updater, cleanup_app,\
    get_global_updater_mngr, set_global_updater_mngr, GlobalUpdaterManager


__all__ = ['get_updater', 'set_updater', 'cleanup_app', 'ThreadUpdater', 'is_main_thread',
           'get_global_updater_mngr', 'set_global_updater_mngr', 'GlobalUpdaterManager']



import threading

import qtpy
from qtpy import QtWidgets

from qt_thread_updater.thread_updater import ThreadUpdater


__all__ = ['get_updater', 'set_updater', 'cleanup_app',
           'get_global_updater_mngr', 'set_global_updater_mngr', 'GlobalUpdaterManager']


def get_updater():
    """Return the main updater. Override `GlobalUpdater.get_updater` to change how this function works.

    Overriding GlobalUpdater.get_updater may be needed if multiprocess/pickling does not work because of the lock.
    """
    return get_global_updater_mngr().get_updater()


def set_updater(updater):
    """Set the main updater. Override `GlobalUpdater.set_updater` to change how this function works."""
    get_global_updater_mngr().set_updater(updater)


def cleanup_app(app=None):
    """Clean up and delete the QApplication singleton."""
    if app is None:
        app = QtWidgets.QApplication.instance()

    # Reset the global reference to the updater
    set_updater(None)

    if app:
        app.closeAllWindows()
        for widg in app.allWindows():
            widg.deleteLater()

    for mod in dir(qtpy):
        try:
            delattr(getattr(qtpy, mod), 'qApp')
        except:
            pass
        try:
            setattr(getattr(qtpy, mod), 'qApp', None)
        except:
            pass


GLOBAL_UPDATER_MANGER = None


def get_global_updater_mngr():
    """Return the manager that creates and returns the ThreadUpdater."""
    global GLOBAL_UPDATER_MANGER
    return GLOBAL_UPDATER_MANGER


def set_global_updater_mngr(mngr):
    """Return the manager that creates and returns the ThreadUpdater."""
    global GLOBAL_UPDATER_MANGER
    GLOBAL_UPDATER_MANGER = mngr


class GlobalUpdaterManager(object):

    def __init__(self):
        self.lock = threading.RLock()
        self.main_updater = ThreadUpdater()

    def get_updater(self):
        """Return the main updater."""
        with self.lock:
            if self.main_updater is None:
                # Create a default updater (may cause issues if this is not created in the main thread)
                self.main_updater = ThreadUpdater(init_later=True)

                # Init later was temporarily used when the lock was not used. Two thread could create the main_updater
                # at the same time causing problems.
                # Initially I did not want to use a lock, because it could break multiprocessing for the module.
                self.main_updater.init()

            return self.main_updater

    def set_updater(self, updater):
        """Set the main updater."""
        with self.lock:
            self.main_updater = updater

    def __getstate__(self):
        """Return the variables to serialize. This should fix multiprocessing with locks pickling problems."""
        # Note do NOT give any lock objects!
        return {}

    def __setstate__(self, state):
        """Initialize this object in a separate process."""
        self.lock = threading.RLock()
        self.main_updater = ThreadUpdater()

        if isinstance(state, dict):
            for k, v in state.items():
                setattr(self, k, v)


# Set the default global updater manager
set_global_updater_mngr(GlobalUpdaterManager())



name = 'qt_thread_updater'
version = '1.1.2'
description = 'Python Qt thread updater to update GUI items using a separate thread.'
url = 'https://github.com/justengel/qt_thread_updater'
author = 'Justin Engel'
author_email = 'jtengel08@gmail.com'



import inspect
import threading
from collections import deque
from qtpy import QtWidgets, QtGui, QtCore
from qt_thread_updater import is_main_thread, get_updater


__all__ = ['clipboard', 'QuickPlainTextEdit', 'QuickTextEdit', 'StreamWrite']


def clipboard(text=None):
    """Return the clipboard text or set the clipboard text if text was given."""
    app = QtWidgets.QApplication.instance()

    if text is None:
        return app.clipboard().text()
    else:
        app.clipboard().setText(text)


class QuickPlainTextEdit(QtWidgets.QPlainTextEdit):
    """QuickPlainTextEdit allows you to quickly write text in a separate thread."""

    DEFAULT_MAX_BLOCKS = 800

    def __init__(self, *args, **kwargs):
        super(QuickPlainTextEdit, self).__init__(*args, **kwargs)

        self._queue_lock = threading.RLock()
        self._queue = deque()

        vert_bar = self.verticalScrollBar()
        self._last_scroll_range = (vert_bar.minimum(), vert_bar.maximum())
        vert_bar.rangeChanged.connect(self._check_scrollToBottom)

        self.setReadOnly(True)
        self.setUndoRedoEnabled(False)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setWordWrapMode(QtGui.QTextOption.WrapAnywhere)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.custom_context_menu)
        self.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard |
                                     QtCore.Qt.TextSelectableByMouse)

        self.setMaximumBlockCount(self.DEFAULT_MAX_BLOCKS)

    def custom_context_menu(self, position):
        """Create the standard context menu for the QTextEdit."""
        menu = self.createStandardContextMenu()

        # Clear
        clear_action = QtWidgets.QAction("Clear", None)
        clear_action.setToolTip("Clear the monitor text.")
        clear_action.triggered.connect(self.clear)
        menu.insertAction(menu.actions()[0], clear_action)

        # Copy all
        copy_all_action = menu.addAction("Copy All")
        # copy_all_action = QtGui.QAction(, None)
        copy_all_action.setToolTip("Copy all of the text in the monitor to the clipboard.")
        copy_all_action.triggered.connect(self.copy_all)
        # menu.addAction(copy_all_action)

        if self.document().isEmpty():
            clear_action.setEnabled(False)
            copy_all_action.setEnabled(False)

        menu.popup(self.mapToGlobal(position))
        return menu
    # end customContextMenu

    def copy_all(self):
        """Copy all of the text in the monitor to the clipboard."""
        clipboard(self.toPlainText())

    def setMaximumBlockCount(self, maximum):
        """Set the maximum number of blocks."""
        with self._queue_lock:
            if maximum <= 0:
                maximum = 0
            self._queue = deque(self._queue, maxlen=maximum)
        super().setMaximumBlockCount(maximum)

    def redirect(self, *iostreams, color=None, fmt=None, **kwargs):
        """Return an object which writes to this text edit with the given color (if supported).

        This is useful for replacing sys.stdout

        ..code-block::python

            widg = QuickTextEdit()
            sys.stderr.write = widg.redirect('red')

        Args:
            *iostreams (object)[None]: Additional object to write to.
            color (str)[None]: String color name to write the text foreground with.
                If fmt argument is not given this defaults to 'black'.
            fmt (QTextCharFormat)[None]: Use this text format to write the text.
                If this is given the color argument will not be used!

        Returns:
            redirect (RedirectStream): Callable write object to change the stdout.write to write to this object.
        """
        return StreamWrite(self, *iostreams, color=color, fmt=fmt, **kwargs)

    def write(self, text, *args, **kwargs):
        """Put data on the queue to add to the TextEdit view."""
        text = str(text)
        if len(text) > 0:
            with self._queue_lock:
                self._queue += text
            get_updater().call_latest(self.update_display)

    def update_display(self):
        """Update the display by taking data off of the queue and displaying it in the widget."""
        with self._queue_lock:
            text = ''.join((self._queue.popleft() for _ in range(len(self._queue))))

        if len(text) == 0:
            return

        cursor = QtGui.QTextCursor(self.textCursor())
        cursor.beginEditBlock()

        # Move and check the position
        # old_pos = cursor.position()
        cursor.movePosition(QtGui.QTextCursor.End)
        # is_end = cursor.position() == old_pos

        # Insert the text (This will insert text even without setTextCursor)
        cursor.insertText(text)
        cursor.endEditBlock()

    def _check_scrollToBottom(self, minv, maxv):
        """When the range changes check if it is larger than the last range and scroll to bottom if it is."""
        if maxv > self._last_scroll_range[1]:
            self._last_scroll_range = (minv, maxv)
            self.scrollToBottom()

    def scrollToBottom(self, *args, **kwargs):
        """Scroll the vertical scroll bar to the maximum position."""
        vert_bar = self.verticalScrollBar()
        vert_bar.setSliderPosition(vert_bar.maximum())


class QuickTextEdit(QtWidgets.QTextEdit):
    """Quick TextEdit for monitoring I/O.

    This works as a redirect for the output stream.

    ..code-block:: python

        import sys
        widg = QuickTextEdit()
        sys.stdout.write = widg.redirect('blue')
        sys.stderr.write = widg.redirect('red')
    """

    DEFAULT_MAX_BLOCKS = 800

    def __init__(self, *args, **kwargs):
        super(QuickTextEdit, self).__init__(*args, **kwargs)

        self._queue_lock = threading.RLock()
        self._queue = deque()

        vert_bar = self.verticalScrollBar()
        self._last_scroll_range = (vert_bar.minimum(), vert_bar.maximum())
        vert_bar.rangeChanged.connect(self._check_scrollToBottom)

        self.setReadOnly(True)
        self.setUndoRedoEnabled(False)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setWordWrapMode(QtGui.QTextOption.WrapAnywhere)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.custom_context_menu)
        self.setTextInteractionFlags(QtCore.Qt.TextSelectableByKeyboard |
                                     QtCore.Qt.TextSelectableByMouse)

        self.setMaximumBlockCount(self.DEFAULT_MAX_BLOCKS)

    def custom_context_menu(self, position):
        """Create the standard context menu for the QTextEdit."""
        menu = self.createStandardContextMenu()

        # Clear
        clear_action = QtWidgets.QAction("Clear", None)
        clear_action.setToolTip("Clear the monitor text.")
        clear_action.triggered.connect(self.clear)
        menu.insertAction(menu.actions()[0], clear_action)

        # Copy all
        copy_all_action = menu.addAction("Copy All")
        # copy_all_action = QtGui.QAction(, None)
        copy_all_action.setToolTip("Copy all of the text in the monitor to the clipboard.")
        copy_all_action.triggered.connect(self.copy_all)
        # menu.addAction(copy_all_action)

        if self.document().isEmpty():
            clear_action.setEnabled(False)
            copy_all_action.setEnabled(False)

        menu.popup(self.mapToGlobal(position))
        return menu

    def copy_all(self):
        """Copy all of the text in the monitor to the clipboard."""
        clipboard(self.toPlainText())

    def setMaximumBlockCount(self, maximum):
        """Set the maximum number of blocks."""
        with self._queue_lock:
            if maximum <= 0:
                maximum = 0
            self._queue = deque(self._queue, maxlen=maximum)
        self.document().setMaximumBlockCount(maximum)

    def redirect(self, *iostreams, color=None, fmt=None, **kwargs):
        """Return an object which writes to this text edit with the given color (if supported).

        This is useful for replacing sys.stdout

        ..code-block::python

            widg = QuickTextEdit()
            sys.stderr.write = widg.redirect('red')

        Args:
            *iostreams (object)[None]: Additional object to write to.
            color (str)[None]: String color name to write the text foreground with.
                If fmt argument is not given this defaults to 'black'.
            fmt (QTextCharFormat)[None]: Use this text format to write the text.
                If this is given the color argument will not be used!

        Returns:
            redirect (RedirectStream): Callable write object to change the stdout.write to write to this object.
        """
        return StreamWrite(self, *iostreams, color=color, fmt=fmt, **kwargs)

    def write(self, text, color=None, fmt=None, *args, **kwargs):
        """Put data on the queue to add to the TextEdit view.

        Args:
            text (str): String text to write.
            color (str)[None]: String color name to write the text foreground with.
                If fmt argument is not given this defaults to 'black'.
            fmt (QTextCharFormat)[None]: Use this text format to write the text.
                If this is given the color argument will not be used!
        """
        text = str(text)
        if len(text) > 0:
            if fmt is None:
                if color is None:
                    color = 'black'
                fmt = self.currentCharFormat()  # QtGui.QTextCharFormat()
                fmt.setForeground(QtGui.QBrush(QtGui.QColor(color)))

            with self._queue_lock:
                self._queue.append((text, fmt))

            get_updater().call_latest(self.update_display)

    def update_display(self):
        """Update the display by taking data off of the queue and displaying it in the widget."""
        with self._queue_lock:
            items = tuple(self._queue.popleft() for _ in range(len(self._queue)))

        if len(items) == 0:
            return

        cursor = QtGui.QTextCursor(self.textCursor())
        cursor.beginEditBlock()

        # Move and check the position
        # old_pos = cursor.position()
        cursor.movePosition(QtGui.QTextCursor.End)
        # is_end = cursor.position() == old_pos

        # Insert the text
        for text, fmt in items:
            cursor.setCharFormat(fmt)
            cursor.insertText(text)

        # End edit
        cursor.endEditBlock()

    def _check_scrollToBottom(self, minv, maxv):
        """When the range changes check if it is larger than the last range and scroll to bottom if it is."""
        if maxv > self._last_scroll_range[1]:
            self._last_scroll_range = (minv, maxv)
            self.scrollToBottom()

    def scrollToBottom(self, *args, **kwargs):
        """Scroll the vertical scroll bar to the maximum position."""
        vert_bar = self.verticalScrollBar()
        vert_bar.setSliderPosition(vert_bar.maximum())


class StreamWrite(object):
    """Custom object to overwrite a stream's write function. This could also be used as an io stream.

    Args:
        *iostreams (tuple/io.StringIO): Any number of io streams to write to.
        color (str)[None]: String color name to write the text foreground with.
            If fmt argument is not given this defaults to 'black'.
        fmt (QTextCharFormat)[None]: Use this text format to write the text.
            If this is given the color argument will not be used!
    """

    def __init__(self, *iostreams, color=None, fmt=None, **kwargs):
        self.iostreams = [stream for stream in iostreams]
        self.color = color
        self.fmt = fmt

    def add_stream(self, stream):
        """Add a stream to write to."""
        self.iostreams.append(stream)

    def remove_stream(self, stream):
        """Remove a stream."""
        try:
            self.iostreams.remove(stream)
        except:
            pass

    def write(self, text, color=None, fmt=None, **kwargs):
        """Write the text to all of the streams.

        This function does not indicate when an error occurs. This function passes kwargs into each stream.write().

        Args:
            text (str): String text to write.
            color (str)[None]: String color name to write the text foreground with.
                If None the set instance value is used. If the set value is also None and fmt is None 'black' is used.
            fmt (QTextCharFormat)[None]: Use this text format to write the text.
                If None the set instance value is used. If this is given the color argument will not be used!
        """
        if fmt is None:
            fmt = self.fmt
        if not isinstance(color, (str, QtGui.QColor)):
            color = self.color
        if isinstance(color, QtGui.QColor):
            color = str(color.name())

        for stream in self.iostreams:
            try:
                sig = inspect.signature(stream.write)
                if 'color' in sig.parameters:
                    stream.write(text, color=color, fmt=fmt, **kwargs)
                else:
                    stream.write(text, **kwargs)
            except (AttributeError, TypeError, ValueError, Exception):
                pass

    def fileno(self):
        """Return the file handler identifier."""
        for stream in self.iostreams:
            try:
                return stream.fileno()
            except (AttributeError, Exception):
                pass
        return -1

    def flush(self):
        """Flush the iostream."""
        for stream in self.iostreams:
            try:
                return stream.flush()
            except (AttributeError, Exception):
                pass

    __call__ = write



"""
Thread Updater module to help update GUI items in a separate thread.
"""
import sys
import time
import threading
import traceback
import contextlib
from collections import OrderedDict

from qtpy import QtCore, QtWidgets


__all__ = ['is_main_thread', 'ThreadUpdater']


def is_main_thread():
    """Return if the current thread is the main thread."""
    return threading.current_thread() is threading.main_thread()


class ThreadUpdater(QtCore.QObject):
    """General timer that will call functions on an interval.

    DEBUG Types:

        * PRINT_ERROR ["print"]: If an error occurs print the traceback to stderr.
        * HIDE_ERROR ["hide"]: If an error occurs ignore it and do not display the error.
        * RAISE_ERROR ["raise"]: If an error occurs actually raise the error. This will crash the updater.

    Args:
         timeout (float/int)[1/30]: Interval in which to run the update functions (in seconds).
         debug_type (str)[PRINT_ERROR]: Control to manage how errors are handled.
         parent (QtCore.QObject)[None]: Parent QObject.
    """

    starting = QtCore.Signal()  # Signal to start the timer in the main thread.
    stopping = QtCore.Signal()  # Signal to stop the timer in the main thread.
    creating = QtCore.Signal()  # Signal to create the timer in the main thread.

    class DebugTypes:
        PRINT_ERROR = 'print'  # Print to stderr
        HIDE_ERROR = 'hide'    # Do nothing and hide the error
        RAISE_ERROR = 'raise'  # Crash the update function and raise the error

    PRINT_ERROR = DebugTypes.PRINT_ERROR
    HIDE_ERROR = DebugTypes.HIDE_ERROR
    RAISE_ERROR = DebugTypes.RAISE_ERROR

    DEFAULT_DEBUG_TYPE = PRINT_ERROR

    def __init__(self, timeout=1/30, debug_type=None, parent=None, init_later=False, **kwargs):
        """Initialize the ThreadUpdater.

        Args:
             timeout (float/int)[1/30]: Interval in which to run the update functions (in seconds).
             debug_type (str)[DEFAULT_DEBUG_TYPE]: Control to manage how errors are handled.
             parent (QtCore.QObject)[None]: Parent QObject.
             init_later (bool)[False]: Manually initialize this object later with `init()`.
        """
        super().__init__(parent)

        if debug_type is None:
            debug_type = self.DEFAULT_DEBUG_TYPE

        # Lock and update variables
        self._lock = threading.RLock()
        self._latest_call = OrderedDict()
        self._every_call = OrderedDict()
        self._always_call = OrderedDict()
        self._delay_call = []

        # Control variables
        self._timeout = timeout
        self.debug_type = debug_type
        self._running = False
        self._tmr = None

        # Try to initialize later so thread variables can be set as fast as possible.
        if not init_later:
            self.init()

    def init(self, *args, **kwargs):
        """Initialize here; Try to make creating the object in __init__ as fast as possible and with little complexity.

        This is to
        reduce the chance that two threads create the global MAIN_UPDATER at the same time. Yes, I've seen this and it
        was problematic.
        """
        # Move to main thread before connecting the signals, so signals run in the main thread
        if not is_main_thread():
            self.moveToThread(QtWidgets.QApplication.instance().thread())

        # Connect the signals
        self.starting.connect(self.start)
        self.stopping.connect(self.stop)
        self.creating.connect(self.create_timer)

        # Create the timer
        self.create_timer()
        return self

    @contextlib.contextmanager
    def handle_error(self, func=None):
        """Context manager to handle exceptions if the unknown update functions cause an error.

        Change how the errors are handled with the "debug_type" variable.
          * PRINT_ERROR ["print"]: If an error occurs print the traceback to stderr.
          * HIDE_ERROR ["hide"]: If an error occurs ignore it and do not display the error.
          * RAISE_ERROR ["raise"]: If an error occurs actually raise the error. This will crash the updater.
        """
        if self.debug_type == self.RAISE_ERROR:
            yield  # If this errors it will crash the updater and raise teh real error.
        elif self.debug_type == self.HIDE_ERROR:
            try:
                yield
            except Exception:
                pass
        else:  # self.debug_type == self.PRINT_ERROR:
            try:
                yield
            except Exception:
                traceback.print_exc()
                print('Error in {}'.format(func.__name__), file=sys.stderr)

    @contextlib.contextmanager
    def restart_on_change(self, restart=None):
        """Context manager to stop and restart the timer if it is running."""
        if restart is None:
            restart = self.is_running()
        if restart:
            self.stop()
            yield
            self.start()
        else:
            yield

    def get_timeout(self):
        """Return the update timer interval in seconds."""
        return self._timeout

    def set_timeout(self, value):
        """Set the update timer interval in seconds."""
        with self.restart_on_change(self.is_running()):
            self._timeout = value
            try:
                self._tmr.setInterval(int(self.get_timeout() * 1000))
            except (AttributeError, RuntimeError, Exception):
                pass

    def create_timer(self):
        """Actually create the timer."""
        # Check to run this function in the main thread.
        if not is_main_thread():
            self.creating.emit()
            return

        self.stop(set_state=False)

        self._tmr = QtCore.QTimer()
        self._tmr.setSingleShot(False)
        self._tmr.setInterval(int(self.get_timeout() * 1000))
        self._tmr.timeout.connect(self.run_update)

    def is_running(self):
        """Return if running."""
        return self._running

    def stop(self, set_state=True):
        """Stop the updater timer."""
        # Check to run this function in the main thread.
        if not is_main_thread():
            self.stopping.emit()
            return

        try:
            self._tmr.stop()
        except:
            pass
        if set_state:
            self._running = False

    def start(self):
        """Start the updater timer."""
        # Check to run this function in the main thread.
        if not is_main_thread():
            self.starting.emit()
            return

        self.stop(set_state=False)
        self._running = True
        if self._tmr is None:
            self.create_timer()  # Should be in main thread
        self._tmr.start()

    def ensure_running(self):
        """If the updater is not running send a safe signal to start it."""
        if not self.is_running():
            self._running = True
            self.starting.emit()

    def register_continuous(self, func, *args, **kwargs):
        """Register a function to be called on every update continuously."""
        with self._lock:
            self._always_call[func] = (args, kwargs)
        self.ensure_running()

    def unregister_continuous(self, func):
        """Unregister a function to be called on every update continuously."""
        with self._lock:
            try:
                self._always_call.pop(func, None)
            except:
                pass

    def call_latest(self, func, *args, **kwargs):
        """Call the most recent values for this function in the main thread on the next update call."""
        with self._lock:
            self._latest_call[func] = (args, kwargs)
        self.ensure_running()

    def now_call_latest(self, func, *args, **kwargs):
        """Call the latest value in the main thread. If this is the main thread call now."""
        if is_main_thread():
            func(*args, **kwargs)
        else:
            self.call_latest(func, *args, **kwargs)

    def call_in_main(self, func, *args, **kwargs):
        """Call this function in the main thread on the next update call."""
        with self._lock:
            try:
                self._every_call[func].append((args, kwargs))
            except (KeyError, IndexError, Exception):
                self._every_call[func] = [(args, kwargs)]
        self.ensure_running()

    def now_call_in_main(self, func, *args, **kwargs):
        """Call in the main thread. If this is the main thread call now."""
        if is_main_thread():
            func(*args, **kwargs)
        else:
            self.call_in_main(func, *args, **kwargs)

    def delay(self, seconds, func, *args, **kwargs):
        """Call the given function after the given number of seconds has passed.

        This will not be accurate unless your timeout is at a high rate (lower timeout number).

        Args:
            seconds (float/int): Number of seconds to wait until calling the function.
            func (callable): Function to call.
            *args (tuple): Positional arguments to pass into the function.
            **kwargs (dict): Keyword arguments to pass into the function.
        """
        now = time.time()  # Note: this is before the lock
        with self._lock:
            self._delay_call.append(DelayedFunc(now, seconds, func, args, kwargs))
        self.ensure_running()

    def run_update(self):
        """Run the stored function calls to update the GUI items in the main thread.

        This function should not be called directly. Call `ThreadUpdater.start()` to run this function on a timer in
        the main thread.
        """
        # Collect the items using the thread safe lock
        with self._lock:
            always = self._always_call.copy()
            latest, self._latest_call = self._latest_call, OrderedDict()
            main, self._every_call = self._every_call, OrderedDict()
            delayed = [self._delay_call.pop(i) for i in reversed(range(len(self._delay_call)))
                       if self._delay_call[i].can_run()]

        # Start running the functions
        for delayed_func in delayed:
            with self.handle_error(delayed_func.func):
                delayed_func.func(*delayed_func.args, **delayed_func.kwargs)

        for func, (args, kwargs) in always.items():
            with self.handle_error(func):
                func(*args, **kwargs)

        for func, (args, kwargs) in latest.items():
            with self.handle_error(func):
                func(*args, **kwargs)

        for func, li in main.items():
            for args, kwargs in li:
                with self.handle_error(func):
                    func(*args, **kwargs)


class DelayedFunc(object):
    def __init__(self, start_time, delay_time, func, args, kwargs):
        self.start_time = start_time
        self.delay_time = delay_time
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def can_run(self):
        """Return if the time to wait is over."""
        return (time.time() - self.start_time) >= self.delay_time

    def wait_for(self, from_time=None):
        """Return the number of seconds from now until this function should run."""
        if from_time is None:
            from_time = time.time()
        wait = (self.start_time + self.delay_time) - from_time
        if wait < 0:
            wait = 0
        return wait
